/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#ifndef __GTK_SCTEXT_H__
#define __GTK_SCTEXT_H__


#include <gdk/gdk.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkeditable.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GTK_TYPE_SCTEXT                  (gtk_sctext_get_type ())
#define GTK_SCTEXT(obj)                  (GTK_CHECK_CAST ((obj), GTK_TYPE_SCTEXT, GtkSCText))
#define GTK_SCTEXT_CLASS(klass)          (GTK_CHECK_CLASS_CAST ((klass), GTK_TYPE_SCTEXT, GtkSCTextClass))
#define GTK_IS_SCTEXT(obj)               (GTK_CHECK_TYPE ((obj), GTK_TYPE_SCTEXT))
#define GTK_IS_SCTEXT_CLASS(klass)       (GTK_CHECK_CLASS_TYPE ((klass), GTK_TYPE_SCTEXT))

typedef struct _GtkSCTextFont       GtkSCTextFont;
typedef struct _GtkSCPropertyMark   GtkSCPropertyMark;
typedef struct _GtkSCText           GtkSCText;
typedef struct _GtkSCTextClass      GtkSCTextClass;

struct _GtkSCPropertyMark
{
  /* Position in list. */
  GList* property;

  /* Offset into that property. */
  guint offset;

  /* Current index. */
  guint index;
};

struct _GtkSCText
{
  GtkEditable editable;

  GdkWindow *text_area;

  GtkAdjustment *hadj;
  GtkAdjustment *vadj;

  GdkGC *gc;

  GdkPixmap* line_wrap_bitmap;
  GdkPixmap* line_arrow_bitmap;

		      /* GAPPED TEXT SEGMENT */

  /* The text, a single segment of text a'la emacs, with a gap
   * where insertion occurs. */
  union { GdkWChar *wc; guchar  *ch; } text;
  /* The allocated length of the text segment. */
  guint text_len;
  /* The gap position, index into address where a char
   * should be inserted. */
  guint gap_position;
  /* The gap size, s.t. *(text + gap_position + gap_size) is
   * the first valid character following the gap. */
  guint gap_size;
  /* The last character position, index into address where a
   * character should be appeneded.  Thus, text_end - gap_size
   * is the length of the actual data. */
  guint text_end;
			/* LINE START CACHE */

  /* A cache of line-start information.  Data is a LineParam*. */
  GList *line_start_cache;
  /* Index to the start of the first visible line. */
  guint first_line_start_index;
  /* The number of pixels cut off of the top line. */
  guint first_cut_pixels;
  /* First visible horizontal pixel. */
  guint first_onscreen_hor_pixel;
  /* First visible vertical pixel. */
  guint first_onscreen_ver_pixel;

			     /* FLAGS */

  /* True iff this buffer is wrapping lines, otherwise it is using a
   * horizontal scrollbar. */
  guint line_wrap : 1;
  guint word_wrap : 1;
 /* If a fontset is supplied for the widget, use_wchar become true,
   * and we use GdkWchar as the encoding of text. */
  guint use_wchar : 1;

  /* Frozen, don't do updates. @@@ fixme */
  guint freeze_count;
			/* TEXT PROPERTIES */

  /* A doubly-linked-list containing TextProperty objects. */
  GList *text_properties;
  /* The end of this list. */
  GList *text_properties_end;
  /* The first node before or on the point along with its offset to
   * the point and the buffer's current point.  This is the only
   * PropertyMark whose index is guaranteed to remain correct
   * following a buffer insertion or deletion. */
  GtkSCPropertyMark point;
  /* Comparison of text property user data. If NULL (default) user
   * data pointers are compared. The function must return true for
   * equal and false for no equal. */
  gboolean (* compare_data) (gpointer, gpointer);
  /* Used for making a copy of user data. If this function is NULL,
   * the pointer is simply copied.  WARNING: This is *not* a good
   * idear if you afterwards delete properies!  The clone and free
   * function should work together, not against each other. */
  gpointer (* clone_data) (gpointer);
  /* Destructor for user data. If NULL g_free is used. */
  void     (* free_data) (gpointer);

			  /* SCRATCH AREA */

  union { GdkWChar *wc; guchar *ch; } scratch_buffer;
  guint   scratch_buffer_len;

			   /* SCROLLING */

  gint last_ver_value;

			     /* CURSOR */

  gint            cursor_pos_x;       /* Position of cursor. */
  gint            cursor_pos_y;       /* Baseline of line cursor is drawn on. */
  GtkSCPropertyMark cursor_mark;        /* Where it is in the buffer. */
  GdkWChar        cursor_char;        /* Character to redraw. */
  gchar           cursor_char_offset; /* Distance from baseline of the font. */
  gint            cursor_virtual_x;   /* Where it would be if it could be. */
  gint            cursor_drawn_level; /* How many people have undrawn. */

			  /* Current Line */

  GList *current_line;

			   /* Tab Stops */

  GList *tab_stops;
  gint default_tab_width;

  GtkSCTextFont *current_font;	/* Text font for current style */

  /* Timer used for auto-scrolling off ends */
  gint timer;
  
  guint button;			/* currently pressed mouse button */
  GdkGC *bg_gc;			/* gc for drawing background pixmap */
};

struct _GtkSCTextClass
{
  GtkEditableClass parent_class;

  void  (*set_scroll_adjustments)   (GtkSCText	    *text,
				     GtkAdjustment  *hadjustment,
				     GtkAdjustment  *vadjustment);
};


GtkType    gtk_sctext_get_type        (void);
GtkWidget* gtk_sctext_new             (GtkAdjustment *hadj,
				     GtkAdjustment *vadj);
void       gtk_sctext_set_editable    (GtkSCText       *text,
				     gboolean       editable);
void       gtk_sctext_set_word_wrap   (GtkSCText       *text,
				     gint           word_wrap);
void       gtk_sctext_set_line_wrap   (GtkSCText       *text,
				     gint           line_wrap);
void       gtk_sctext_set_adjustments (GtkSCText       *text,
				     GtkAdjustment *hadj,
				     GtkAdjustment *vadj);
void       gtk_sctext_set_point       (GtkSCText       *text,
				     guint          index);
void       gtk_sctext_set_property_data_functions (GtkSCText *text,
				     gboolean (* compare_data) (gpointer,
								gpointer),
				     gpointer (* clone_data) (gpointer),
				     void     (* free_data) (gpointer));
guint      gtk_sctext_get_point       (GtkSCText       *text);
guint      gtk_sctext_get_length      (GtkSCText       *text);
void       gtk_sctext_freeze          (GtkSCText       *text);
void       gtk_sctext_thaw            (GtkSCText       *text);
void       gtk_sctext_insert          (GtkSCText       *text,
				     GdkFont       *font,
				     GdkColor      *fore,
				     GdkColor      *back,
				     const char    *chars,
				     gint           length);
void       gtk_sctext_insert_with_data (GtkSCText       *text,
				     GdkFont       *font,
				     GdkColor      *fore,
				     GdkColor      *back,
				     gpointer       data,
				     const char    *chars,
				     gint           length);
gint       gtk_sctext_backward_delete (GtkSCText       *text,
				     guint          nchars);
gint       gtk_sctext_forward_delete  (GtkSCText       *text,
				     guint          nchars);
void       gtk_sctext_set_property    (GtkSCText       *text,
				     guint          from,
				     guint          to,
				     GdkFont       *font,
				     GdkColor      *fore,
				     GdkColor      *back,
				     gpointer       data);
void       gtk_sctext_set_property_data (GtkSCText     *text,
				       guint        pos,
				       gpointer     user_data);
GdkFont   *gtk_sctext_get_property_font (GtkSCText     *text,
				       guint        pos);
GdkColor  *gtk_sctext_get_property_fore (GtkSCText     *text,
				       guint        pos);
GdkColor  *gtk_sctext_get_property_back (GtkSCText     *text,
				       guint        pos);
guint      gtk_sctext_get_property_length (GtkSCText   *text,
					 guint      pos);
gpointer   gtk_sctext_get_property_data (GtkSCText     *text,
				       guint        pos);

#define GTK_SCTEXT_INDEX(t, index)	(((t)->use_wchar) \
	? ((index) < (t)->gap_position ? (t)->text.wc[index] : \
					(t)->text.wc[(index)+(t)->gap_size]) \
	: ((index) < (t)->gap_position ? (t)->text.ch[index] : \
					(t)->text.ch[(index)+(t)->gap_size]))

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GTK_SCTEXT_H__ */
